Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq fintype.
Require Import bigop ssralg zmodp matrix ssrnum complex.
Require Import big_minmax.

(******************************************************************************)
(*                                                                            *)
(* Here we define a partial order over matrices.                              *)
(*                                                                            *)
(*                                                                            *)
(*         A <m: B == forall i and j, A i j < B i j.                          *)
(*        A <=m: B == forall i and j, A i j <= B i j.                         *)
(*          Mabs A == the matrix where each coefficients is the               *)
(*                    absolute value (or module) of each coefficients         *)
(*                    of the atrix A.                                         *)
(******************************************************************************)


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.
 
Local Open Scope ring_scope.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Section matRDef.

Variable R : numDomainType.
Variable p q: nat.
 
(** Definitions for absolute value of a matrix and componentwise order *)

Definition Mabs (A: 'M[R]_(p, q)) := map_mx normr A.

Definition Mle (A B: 'M[R]_(p, q)) := forall i j, A i j <= B i j. 

Definition Mlt (A B: 'M[R]_(p, q)) := forall i j, A i j < B i j.

Definition one_M: matrix R p q := \matrix_(i, j) 1.

End matRDef.

Implicit Arguments one_M[p q].

Infix "<m:" := Mlt (at level 70): ring_scope.
Infix "<=m:" := Mle (at level 70): ring_scope.

Section matRProp.

Variable R : numDomainType.
Variable n p q: nat.

Open Scope ring_scope.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Lemma Mle_trans (A B C : 'M[R]_(p, q)) : A <=m: B -> B <=m: C -> A <=m: C.
Proof. by move=> HAB HBC i j; rewrite (ler_trans (HAB i j)). Qed.

Lemma Mle_refl (A : 'M[R]_(p, q)) : A <=m: A.
Proof. by []. Qed.

Lemma Mlt_trans (A B C : 'M[R]_(p, q)) : A <m: B -> B <m: C -> A <m: C.
Proof. by move=> HAB HBC i j; rewrite (ltr_trans (HAB i j)). Qed.

Lemma Mlt_le_trans (A B C : 'M[R]_(p, q))  : A <m: B -> B <=m: C -> A <m: C.
Proof. by move=> HAB HBC i j; rewrite (ltr_le_trans (HAB i j)). Qed.

Lemma MltW (A B : 'M[R]_(p, q)) : A <m: B -> A <=m: B.
Proof. by move=> HAB i j; rewrite ltrW. Qed.

Lemma Mabs_eq0 (A :  'M[R]_(p, q)) : ((Mabs A) == 0) = (A == 0).
Proof. 
apply/idP/idP=> [/eqP/matrixP H|/eqP ->]; apply/eqP/matrixP=> i j; rewrite !mxE.
  apply/eqP; rewrite -normr_eq0.
  by have/(_ i j) := H; rewrite !mxE=> ->.
by rewrite normr0.
Qed.

Lemma Mabs_ge0 (A :  'M[R]_(p, q)) : 0 <=m: Mabs A.
Proof. by move=> i j; rewrite !mxE normr_ge0. Qed.

Lemma Mabs_triang (A B : 'M[R]_(p,q)) : (Mabs (A + B)) <=m: (Mabs A + Mabs B).
Proof. by move=> i k; rewrite !mxE ler_norm_add. Qed.

Lemma Mle_add:
forall (a b c d: 'M[R]_(p,q)), Mle a b -> Mle c d -> Mle (a + c) (b + d).
Proof. by move=> a b c d Hab Hcd i k; rewrite !mxE ler_add. Qed.

Lemma MabsM (M : 'M[R]_(n, p)) (N :'M_(p, q)) :
  Mabs (M *m N) <=m: (Mabs M) *m (Mabs N).
Proof.
move=> i k; rewrite !mxE (ler_trans (ler_norm_sum _ _ _)) //.
by apply: ler_sum=> j _; rewrite normrM !mxE. 
Qed.

Lemma Mabs_scalemx  a (M : 'M[R]_(n, p)) : Mabs (a *: M) = `|a| *: Mabs M.
Proof. by apply/matrixP=> i j; rewrite !mxE normrM. Qed.

Lemma eq_Mabs_id (M : 'M[R]_(n, p)) : 0 <=m: M -> Mabs M = M.
Proof. 
move=> Hge0; apply/matrixP=> i j; rewrite mxE.
apply/eqP; rewrite eqr_norm_id.
by have/(_ i j) := Hge0; rewrite mxE.
Qed.

Lemma subr_Mge0 (A B : 'M[R]_(p,q)) : B <=m: A <-> 0 <=m: (A - B).
Proof. 
split=> HAB i j; rewrite ?mxE ?subr_ge0 //.
by have/(_ i j) := HAB; rewrite !mxE subr_ge0.
Qed.

Lemma subr_Mgt (A B C : 'M[R]_(p,q)) : B + A <m: C <-> A <m: C - B.
Proof. by split=> H i j; have/(_ i j):= H; rewrite !mxE ltr_subr_addl. Qed.

Lemma mulmx_gt0 (A: 'M[R]_(p,q)) (x : 'cV_q) : 0 <m: A -> 
   0 <=m: x -> x != 0 -> 0 <m: A *m x.
Proof.
case: q A x=>[A x|m A x].
  by have ->: x == 0 by apply/eqP/matrixP=> [][][].  
move => Apos xpos xneq0 i j; rewrite !mxE.
pose k := odflt ord0 [pick k | x k j != 0].
rewrite (bigD1 k) //= ltr_spaddl // ?mulr_gt0 // ?[_ < x _ _]ltr_def.
    by have/(_ i k) := Apos; rewrite mxE. 
  have/(_ k j):= xpos; rewrite mxE => ->.
  rewrite andbT /k; case: pickP=> //= predx.
  rewrite predx -(negbTE xneq0).
  apply/eqP/matrixP=> h l; rewrite mxE; apply/eqP.
  by rewrite /is_true -[true]negbK [X in ~~ X]/= -(predx h) !ord1 negbK.
apply: sumr_ge0=> l _; rewrite mulr_ge0 //.
  by have/(_ i l):= Apos; rewrite mxE; apply: ltrW.
by have/(_ l j):= xpos; rewrite mxE.
Qed.

Lemma Mlt_scaler (A B : 'M[R]_(p,q)) a : 0 < a -> A <m: B -> a *: A <m: a *: B.
Proof. by move=> Hagt0 HAltB i j; rewrite !mxE ltr_pmul2l. Qed.

Lemma ltr_sum (I : eqType) (r : seq I) (P : pred I) (F G : I -> R) (i : I) :
 (forall i, P i -> F i < G i) -> i \in r -> P i ->
 \sum_(i <- r | P i) F i < \sum_(i <- r | P i) G i.
Proof.
move=> Hlt.
elim: r=> // a l IHl.
rewrite in_cons=> /orP []=> [/eqP -> Ha|Hi HPi].
  rewrite !big_cons Ha ltr_le_add // ?Hlt // ler_sum // => j Hj.
  by rewrite ltrW // Hlt //.
rewrite !big_cons; case H: (P a); last exact: IHl.
by rewrite ltr_add ?Hlt // IHl. 
Qed.

Lemma Mlt_pmul2r (A B : 'M[R]_(n, p.+1)) (C :'M_(p.+1, q)) : 
  0 <m: C ->  A <m: B -> A *m C <m: B *m C. 
Proof.
move=> HCgt0 HAltB i j; rewrite !mxE.
apply: (ltr_sum (i:= ord0))=> // [k _|]; last by rewrite mem_index_enum.
have/(_ k j):= HCgt0; rewrite mxE=> HC.
have/(_ i k) HA:= HAltB.
by rewrite ltr_pmul2r. 
Qed.

Lemma Mlt_pmul2l (C : 'M[R]_(n, p.+1)) (A B :'M_(p.+1, q)) : 
  0 <m: C ->  A <m: B -> C *m A <m: C *m B. 
Proof.
move=> HCgt0 HAltB i j; rewrite !mxE.
apply: (ltr_sum (i:= ord0))=> // [k _|]; last by rewrite mem_index_enum.
have/(_ i k):= HCgt0; rewrite mxE=> HC.
have/(_ k j) HA:= HAltB.
by rewrite ltr_pmul2l. 
Qed.

Lemma Mle_wpmul2l (C : 'M[R]_(n, p)) (A B :'M_(p, q)) : 
  0 <=m: C ->  A <=m: B -> C *m A <=m: C *m B. 
Proof.
move=> HCgt0 HAltB i j; rewrite !mxE.
apply: (ler_sum )=> // k _.
have/(_ i k):= HCgt0; rewrite mxE=> HC.
have/(_ k j) HA:= HAltB.
by rewrite ler_wpmul2l. 
Qed.

Lemma Mle_wpmul2r (C : 'M[R]_(p, n)) (A B :'M_(q, p)) : 
  0 <=m: C ->  A <=m: B -> A *m C <=m: B *m C. 
Proof.
move=> HCge0 HAleB i j; rewrite !mxE.
apply: (ler_sum )=> // k _.
have/(_ k j):= HCge0; rewrite mxE=> HC.
have/(_ i k) HA:= HAleB.
by rewrite ler_wpmul2r.
Qed.

End matRProp.

Section matRProp2.
Variable R : rcfType.
Variable p q n : nat.

Lemma Mle_wpmul :
forall (a b: 'M[R]_(q,n)) (u v: 'M_(p, q)),
  u <=m: v -> a <=m: b -> Mle 0 a -> Mle 0 u -> u *m a <=m: v *m b.
move=> a b u v Huv Hab Ha Hu.
have Hb := Mle_trans Ha Hab.
apply: Mle_trans (Mle_wpmul2r Hb Huv).
exact: Mle_wpmul2l.
Qed.

Local Notation "A ^%:C" := (map_mx (real_complex_def (Phant R)) A) (at level 2).

Lemma Mabs_real_complex (A : 'M[R]_(p,q)) : Mabs A^%:C = (Mabs A)^%:C. 
Proof.
apply/matrixP=> i j; rewrite !mxE normcE sqrtc_sqrtr ?mulcJ_ge0 //.
by rewrite conjc_real -rmorphM -sqrtr_sqr.
Qed.

Lemma mxreal_gt0 m (A : 'M[R]_(m,n)) : 0 <m: A <-> 0 <m: A^%:C.
Proof. by split=> Hpos i j; have/(_ i j) := Hpos; rewrite !mxE ltcR. Qed.

Lemma mxreal_ge0 m (A : 'M[R]_(m,n)) : 0 <=m: A <-> 0 <=m: A^%:C.
Proof. by split=> Hpos i j; have/(_ i j) := Hpos; rewrite !mxE lecR. Qed.

End matRProp2.

Section theory.

Variable R : numFieldType.

Lemma Mle_eps m (u v : 'cV[R]_m.+1) :
    let F i := ((u i 0) / 2%:R)/ (v i 0) in  
    let eps := min_fT F in
    0 <m: u -> 0 <m: v -> eps *: v <m: u.  
Proof.
move=> F eps Hu Hv i j; rewrite ord1 !mxE.
have HumxE l : 0 < u l 0 by have/(_ l 0):= Hu; rewrite mxE. 
have HvmxE l : 0 < v l 0 by have/(_ l 0):= Hv; rewrite mxE. 
have [k Hk] := ex_min_fT F ord0.
have unitv l : v l 0 \is a GRing.unit.
  by have:= Hv l 0; rewrite mxE lt0r unitfE => /andP [].  
apply: (ler_lt_trans (y:= u i 0 / 2%:R)).
  rewrite /eps Hk /F -[u i _ / _]mulr1 -[X in _ <= _ * X](mulVr (unitv i)).
  rewrite mulrA ler_pmul ?mulr_ge0 ?invr_ge0 ?[0 <= _]ltrW ?ltr0n //.
  rewrite /F in Hk; rewrite -Hk  min_fT_le // => l.
  by rewrite gtr0_real // ?mulr_gt0 // invr_gt0 // ltr0n.
by rewrite mulrC gtr_pmull // invr_lt1 ?ltr0n ?ltr1n // unitfE pnatr_eq0.
Qed.

End theory.


