Require Import Classical_Prop Classical_Pred_Type.
Require Import IndefiniteDescription ChoiceFacts.
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq.

(******************************************************************************)
(*                                                                            *)
(*   This file contains the set theory that will be use for the devloppment   *)
(*   of topology. A set is represented by this caracteristic proposition,     *)
(*   that is a set of elemnts of type T has type T -> Porp.                   *)
(*                                                                            *)
(*         If H has type {x : T | P x} then                                   *)
(*             pi1 H == x. It is a shortcut for proj1_sig.                    *)
(*             pi2 H == a proof of P x. It is a shortcut for proj2_sig.       *)
(*         If H has type exists x : T, P x then                               *)
(*             cid H == an inhabitant of type {x : T | P x}. It is a          *)
(*                      shortcut for constructive_indefinite_description.     *)
(*                                                                            *)
(*  *** Basic definitions                                                     *)
(*           {ø : T} == the empty set over element of type T.                 *)
(*        {full : T} == the set that contains all elements of type T.         *)
(*           x 'in A == x is an element of the set A.                         *)
(*                        (i.e it is a proof of A x).                         *)
(*            A =s B == A is the same set of B.                               *)
(*            A <s B == the set A is include in the set B.                    *)
(*               A^c == the absolute complement of A.                         *)
(*             A \ B == the set A private of the elements of the set B.       *)
(*                      (it is the relative complement of B in A).            *)
(*         inter A B == the intersection of A and B.                          *)
(*         union A B == the union of A and B.                                 *)
(*       inter_s E F == if E is a set of set (i.e the type                    *)
(*                      of E is (T -> Prop) -> Prop), it is the               *)
(*                      intersection of the sets F A such that A 'in E.       *)
(*       union_s E F == if E is a set of set (i.e the type                    *)
(*                      of E is (T -> Prop) -> Prop), it is the               *)
(*                        union of the sets F A such that A 'in E.            *)
(*                                                                            *)
(*  *** Functions                                                             *)
(*           image f == the set of y such that exists x, y = f x.             *)
(*             f^(U) == the set of f x such that x 'in U.                     *)
(*            f^-1 y == the set of elements x such that y = f x.              *)
(*         f@^-1 (U) == the union of f^-1 y such that y 'in U.                *)
(*       inverse Hbf == if Hbf is a proof that f is bijective, it is          *)
(*                      the inverse function of f. The pretty printer uses    *)
(*                      the notation f-1.                                     *)
(*                                                                            *)
(*  *** Family of sets                                                        *)
(*      finite_set A == A is a finite set.                                    *)
(*       family T Ti == the type of family of sets (of type T -> Prop)        *)
(*                      where Ti is the type of elements that indexies        *)
(*                      the family.                                           *)
(*           (ind f) == the index of the family of set f.                     *)
(*     finite_family == the type of finite family of sets.                    *)
(*     cardinality f == the number of sets in the family f.                   *)
(*    sub_family f g == the family of set g is a subfamily of f.              *)
(*         ind_seq f == the sequence of all index of finite family f.         *)
(*       empty_fam f == f is an empty family of set.                          *)
(*            fempty == the empty family of set.                              *)
(*       inter_fam f == the intersection of all set in the family f.          *)
(*       union_fam f == the union of all set in the family f.                 *)
(*         cover A f == the family of set f cover the set A.                  *)
(*  fam_fin_to_seq f == the sequence that contains all set of                 *)
(*                      finite family f (it is of type seq (T -> Prop)).      *)
(*       inter_seq s == the intersection of all set in the sequence s.        *)
(*                      we define this to allows proof by induction in        *)
(*                      the case that we have finite number of sets.          *)
(******************************************************************************)


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


Notation pi1 := proj1_sig.
Notation pi2 := proj2_sig.
Notation cid := (constructive_indefinite_description _).
 
Section Teqtype.

Lemma cddo : forall T : Type, ConstructiveDefiniteDescription_on T.
Proof.
by move=> T P H; apply: cid; case : H=> x [Hx H2]; exists x.
Qed.

Lemma eqProp : forall (P : Prop), {P} + {~P}.
Proof.
apply: (constructive_definite_descr_excluded_middle cddo)=> P.
exact: (classic P).
Qed.

Definition Propb (P : Prop) : bool:=
  match (eqProp P) with
    | left _ => true
    | right _ => false
  end.

Lemma Pb : forall (P : Prop), reflect P (Propb P). 
Proof.
move=> P.
apply: (iffP idP)=> H; rewrite /Propb; case: (eqProp P)=> // Ha.
have Hab : Propb P = false by rewrite /Propb; case: (eqProp P).
by rewrite H in Hab.
Qed.

 Variable T : Type.

Definition eqT (x y : T) : bool := Propb (x = y).

Lemma eqTP : Equality.axiom eqT.
Proof.
by move=> x y; apply: Pb.
Qed.

Canonical Structure T_eqMixin := Eval hnf in EqMixin eqTP.

End Teqtype.

Section prelemme.

Lemma orPC (A B : Prop) : (A \/ B) -> (B \/ A).
Proof. by  case; [apply: or_intror|apply: or_introl]. Qed.

Lemma andPC (A B : Prop) : (A /\ B) <-> (B /\ A).
Proof. by split; case. Qed.

Lemma not_and_to_imply (A B : Prop) : ~(A /\ B) <-> (B -> ~A).
Proof.
by split=> H; 
 [apply: or_to_imply; apply/orPC; apply: not_and_or 
 | apply: or_not_and; apply/orPC; apply: imply_to_or].
Qed.

Lemma not_imply_to_and (A B : Prop) : ~(A -> B) <-> (A /\ ~ B).
Proof.
split=> [|[HA HB] Habs]; first exact (imply_to_and A B).
exact: (HB (Habs HA)).
Qed.

End prelemme.

Section Ensemble.

Variable T : Type.

Require Export Setoid.
Require Export Morphisms.

Definition in_set (x : T) (A : T-> Prop) := A x.
Notation "x 'in A" := (in_set x A) (at level 3).

Lemma insE x (P : T -> Prop) : x 'in (fun x => P x) = P x.
Proof. done. Qed.

Definition eqset (A B : T -> Prop) : Prop := forall x, x 'in A  <-> x 'in B.
Notation "A =s B" := (eqset A B) (at level 70, no associativity).

Lemma eqset_refl : forall A, A =s A.
Proof. done. Qed.

Lemma eqset_sym : forall A B, A =s B -> B =s A.
Proof. by move=> A B H x; case: (H x). Qed.

Lemma eqset_trans : forall A B C, A =s B -> B =s C -> A =s C.
Proof.
move=> A B C HAB HBC; split=> H.
  by apply/HBC/HAB.
by apply/HAB/HBC. 
Qed.

Add Relation  _ eqset
reflexivity proved by eqset_refl
symmetry proved by eqset_sym
transitivity proved by eqset_trans
as eqset_rel. 


Add Parametric Morphism x : (in_set x) 
  with signature eqset ==> iff as in_set_mor.
Proof. by move=> A B HAB; split=> /HAB. Qed.

Definition empty (x : T) : Prop := False.

Definition full (x : T) : Prop := True.

Definition include (A B : T -> Prop) := forall (x: T), x 'in A -> x 'in B.
Notation "A <s B" := (include A B) (at level 70, no associativity).

Add Morphism include with signature eqset ==> eqset ==> iff as include_mor.
Proof. by move=> A B HAB C D HCD; split=> H x /HAB/H/HCD. Qed.

Lemma include_trans A B C : A <s B -> B <s C -> A <s C.
Proof. by move=> HAB HBC x /HAB /HBC. Qed.

Lemma left_i A B : A =s B -> A <s B.
Proof. by move=> ->. Qed.

Lemma right_i A B : A =s B -> B <s A.
Proof. by move=> ->. Qed.

Lemma double_i A B : A <s B -> B <s A -> A =s B.
Proof. by move=> HAB HBA; split; move: x. Qed.

Definition complementary  (A : T -> Prop) x : Prop := ~ x 'in A.
Notation "A ^c" := (complementary A) (at level 2, no associativity).

Add Morphism  complementary with signature eqset ==> eqset as compl_mor.
Proof. by move=> A B HAB; split=> /HAB. Qed.

Lemma compl_invo A : (A^c)^c =s A.
Proof. by split; [apply: NNPP |move=> ?]. Qed.

Lemma compl_include A B : B^c <s A^c <-> A <s B.
Proof. by split=> H; [rewrite -[B]compl_invo|]=> x Hx /H. Qed.

Lemma empty_c : empty^c =s full.
Proof. by split=> // _. Qed.

Lemma full_c : full^c =s empty.
Proof. by rewrite -empty_c compl_invo. Qed.

Lemma neq_compl A x y : x 'in A^c  -> y 'in A -> x <> y.
Proof. by move=> Hx Hy Habs; rewrite Habs in Hx. Qed.

Definition inter (A B: T -> Prop) x :Prop := x 'in A /\ x 'in B.

Add  Morphism inter with signature eqset ==> eqset ==> eqset as inter_mor. 
Proof. by move=> A B HAB C D HCD; split=> [][/HAB H1 /HCD H2]. Qed.

Lemma inter_sym A B : inter A B =s inter B A.
Proof. by split=> [][]. Qed.

Lemma inter_neutr A B : A <s B -> inter A B =s A.
Proof. by move=> HAB; split=> [[]|H] //; split=> //; apply: HAB. Qed.

Lemma inter_left A B : inter A B <s A.
Proof. by move=> x []. Qed.

Lemma inter_right A B : inter A B <s B.
Proof. by move=> x []. Qed.

Lemma in_inter A B C D : A <s C -> B <s D -> inter A B <s inter C D.
Proof. by move=> HAC HBD x [/HAC H1 /HBD H2]. Qed.

Definition union (A B : T -> Prop) x := x 'in A \/ x 'in B.

Add Morphism union with signature eqset ==> eqset ==> eqset as union_mor.
Proof.
by move=> A B HAB C D HCD; split; case=> [/HAB|/HCD] H;[left|right|left|right].
Qed.

Lemma union_sym  A B : union A B =s union B A.
Proof. by split=>/orPC. Qed.

Lemma union_neutr A B : B <s A -> union A B =s A.
Proof. by move=> HAB; split=> [[] // /HAB //|?]; left. Qed.

Lemma union_left A B : A <s union A B.
Proof. by move=> x; left. Qed.

Lemma union_right A B : B <s union A B.
Proof. by move=> x; right. Qed.

Lemma in_union A B C D : A <s C -> B <s D -> union A B <s union C D.
Proof. by move=> HAC HBD x; case=>[/HAC H|/HBD H]; [left|right]. Qed. 

Lemma compl_inter A B : (inter A B)^c =s union A^c B^c.
Proof. by split=> H; [apply: not_and_or|apply: or_not_and]. Qed.

Lemma compl_union A B : (union A B)^c =s inter A^c B^c.
Proof. by split => H; [apply: not_or_and|apply: and_not_or]. Qed.

Definition without (A B : T-> Prop) := inter A B^c.

Notation "A \ B" := (without A B) (at level 0, no associativity).

Add Morphism without with signature eqset ==> eqset ==> eqset as without_mor.
Proof. by move=> A B HAB C D; rewrite /without HAB=> ->. Qed.

Lemma A_without_A A : A\A =s empty.
Proof. by split=> // [][]. Qed.

End Ensemble.

Existing Instance eqset_rel.
Existing Instance in_set_mor_Proper.
Existing Instance include_mor_Proper.
Existing Instance compl_mor_Proper.
Existing Instance inter_mor_Proper.
Existing Instance union_mor_Proper.
Existing Instance without_mor_Proper.

Notation "x 'in A" := (in_set x A) (at level 3).
Notation "A =s B" := (eqset A B) (at level 70, no associativity).
Notation "A <s B" := (include A B) (at level 70, no associativity).
Notation "A ^c" := (complementary A) (at level 2, no associativity).
Notation "A \ B" := (without A B) (at level 0, no associativity).
Notation "{ø : T }" := (@empty T).
Notation "{full : T }" := (@full T).

Add Parametric Morphism T : (@in_set (T -> Prop)) 
  with signature (@eqset T) ==> ((@eqset T) ==> iff) ==> iff as in_setl_mor.
Proof. by move=> A B HAB C D HCD; split=> H; apply/(HCD A B). Qed.


Section inter_union.

Variable T : Type.

Definition inter_s (E : (T -> Prop) -> Prop) (F : (T-> Prop) -> (T -> Prop)) 
  (x : T) := forall A, A 'in E -> x 'in (F A) .

Definition union_s (E : (T -> Prop) -> Prop) (F : (T-> Prop) -> (T -> Prop)) 
  (x : T) := exists A, A 'in E /\ x 'in (F A).

Lemma inters_in (E : (T -> Prop) -> Prop) F A : 
  A 'in E -> inter_s E F <s F A.
Proof. by move=> H x; apply. Qed.

Lemma in_unions (E : (T -> Prop) -> Prop) F A : 
  A 'in E -> F A <s union_s E F.
Proof. by move=> H x Hx; exists A. Qed.

Lemma inters_in2 E F A : A <s inter_s E F <-> forall S, S 'in E -> A <s F S.
Proof.
split=> [H S /inters_in HS|H x Hx S /H]; last by apply. 
exact: (include_trans H). 
Qed.

Lemma in_unions2 E F A : union_s E F <s A <-> forall S, S 'in E -> F S <s A.
Proof.
split=> [H S /in_unions HS|H x [S [/H]]]; last by apply.
exact: (include_trans _ H).
Qed.

Lemma union_compl_inter E F :
  (union_s E F)^c =s (inter_s E (fun A => (F A)^c)).
Proof.
split=> [/not_ex_all_not H A|H]; first by apply/not_and_to_imply/andPC.
by apply: all_not_not_ex => A; apply/andPC/not_and_to_imply/H.
Qed.

Lemma inter_compl_union E F : 
  (inter_s E F)^c =s (union_s E (fun A => (F A)^c)). 
Proof.
split; first by case/not_all_ex_not=> A HA; exists A; apply: imply_to_and.
by case=> A HA; apply: ex_not_not_all; exists A; apply/not_imply_to_and.
Qed.

Lemma eq_union E F : 
union_s E F =s union_s (fun B=> exists A, A 'in E /\ (F A =s B)) id. 
Proof.
apply/double_i/in_unions2=> x [A [HA1 HA]].
  by exists (F A); split=> //; exists A.
by exists A; rewrite HA.
Qed.

Lemma eq_inter E F :
inter_s E F =s inter_s (fun B=> exists A, A 'in E /\ (F A =s B)) id.
Proof.
apply/double_i/inters_in2 => [x Hx A [S [/Hx HS <-]]|S HS A] //.
by apply; exists S.
Qed.

Lemma eq_union2 E F E' F' : E =s E' -> (forall A, F A =s F' A) ->
  union_s E F =s union_s E' F'.
Proof. by move=> HE HG x; split; case=> S [/HE HS1 /HG HS2]; exists S. Qed.

Lemma eq_inter2 E F E' F' : E =s E' -> (forall A, F A =s F' A) ->
  inter_s E F =s inter_s E' F'.
Proof. by move=> HE HG x; split=> Hx A /HE /Hx /HG HA. Qed.

End inter_union.

Section function.

Variables T1 T2 T3 : Type.

Definition image (f : T1 -> T2) y := exists x, y = f x.

Definition range (f : T1 -> T2) (U : T1 -> Prop) (y : T2) :=
  exists x, x 'in U /\ y = f x.
Notation "f ^( U )" := (range f U) (at level 0).

Add Parametric Morphism (f : T1 -> T2) :
  (range f) with signature (@eqset T1) ==> (@eqset T2) as rng_mor.
Proof.
move=> A B HAB.
by apply: double_i=> x ; case=> y [H1 H2]; exists y ; split=> // ; apply/HAB.
Qed.

Definition fiber (f : T1 -> T2) (y : T2) (x : T1) := y = f x.
Notation "f ^-1" := (fiber f).

Definition preimage (f : T1 -> T2) (U : T2 -> Prop) (x : T1) :=
   exists y, y 'in U /\ x 'in (f^-1 y).
Notation "f @^-1( U )" := (preimage f U) (at level 0).

Add Parametric Morphism (f : T1 -> T2) :
  (preimage f) with signature (@eqset T2) ==> (@eqset T1) as preim_mor.
Proof.
move=> A B HAB.
by apply: double_i=> y; case=> x [H1 H2]; exists x; split=> //; apply/HAB.
Qed.

Definition injective (f : T1 -> T2) := forall x y, f x = f y -> x = y.
Definition surjective (f : T1 -> T2) := forall (y : T2), exists x, y = f x.
Definition bijective (f : T1 -> T2) := injective f /\ surjective f.

Definition inverse f (Hbf : bijective f) := fun y => pi1 (cid ((proj2 Hbf) y)).

Lemma surj_eq : forall f, surjective f <-> f^({full : T1}) =s {full : T2}.
Proof.
move=> f; split=> H y.
  by split=> Hy //; case: (H y)=> x Hx; exists x.
by case: (H y)=> _; case=> // x Hx; exists x; case: Hx.
Qed.

Lemma in_finv_f : forall (f : T1 -> T2) (U : T1 -> Prop), U <s f@^-1(f^(U)).
Proof. by move=> f U x Hx; exists (f x); split=> //; exists x. Qed.

Lemma f_finv_in : forall (f : T1 -> T2) (U : T2 -> Prop), f^(f@^-1(U)) <s U.
Proof. by move=> f U x; case=> y; case; case=> z [H1 <-] ->. Qed.

Lemma surj_invo : forall f U, surjective f -> f^(f@^-1(U)) =s U.
Proof.
move=> f U H; apply: double_i; first by apply: f_finv_in.
by move=> x Hx; case: (H x)=> y Hy; exists y; split=> //; exists x.
Qed.

Lemma inj_invo : forall f U, injective f -> f@^-1(f^(U)) =s U.
Proof.
move=> f U H; apply: double_i; last by apply: in_finv_f.
move=> x; case=> y; case; case=> z [H1 H2] Hy.
by rewrite Hy in H2; rewrite (H _ _ H2).
Qed.

Lemma compl_finv : forall f U, f@^-1(U^c) =s f@^-1(U)^c.
Proof.
move=> f U; apply: double_i=> x.
  case=> y [H1 H2].
  rewrite insE in H2.
  apply: all_not_not_ex=> z; case; rewrite !insE -H2 => Hz Hz2.
  by rewrite Hz2 in Hz.
move=> Hx; move: (not_ex_all_not _ _ Hx) => H.
by exists (f x); split=> // Habs; case: (H (f x)).
Qed.

Lemma compl_f_inj : forall f U, injective f ->  f^(U^c) <s f^(U)^c.
Proof.
move=> f U Hf x; case=> y [Hy1 Hy2]; apply: all_not_not_ex=> z; case.
rewrite Hy2=> Hz1 Hz2.
by rewrite (Hf _ _ Hz2) in Hy1.
Qed.

Lemma compl_f_surj : forall f U, surjective f -> f^(U)^c <s f^(U^c).
Proof.
move=> f U Hf y Hy; case: (Hf y)=> z Hz; exists z; split=> // Habs.
by move/not_ex_all_not: Hy => H; case: (H z).
Qed.

Lemma compl_f_bij : forall f U, bijective f -> f^(U^c) =s f^(U)^c.
Proof.
move=> f U [Hf1 Hf2]; apply: double_i;
 [exact (compl_f_inj Hf1)
 | exact (compl_f_surj Hf2)].
Qed.

End function.

Existing Instance rng_mor_Proper.
Existing Instance preim_mor_Proper.

Notation "f ^( U )" := (range f U) (at level 0).
Notation "f ^-1" := (fiber f).
Notation "f @^-1( U )" := (preimage f U) (at level 0).
Notation "f '-1'" := (@inverse _ _ f _)(at level 30, format "f '-1'"). 

Section function2.

Variables T1 T2 T3 : Type.

Lemma comp_inv : forall (f : T1 -> T2) (g : T3 -> T1) U,
   (f \o g)@^-1(U) =s g@^-1(f@^-1(U)).
Proof.
move=> f g U; apply: double_i=> x; case=> y [Hy1 Hy2].
  by exists (g x); split=>//; exists y.
by case: Hy1=> z [Hz1 Hz2]; exists z; split=> //; rewrite Hz2 Hy2.
Qed.

End function2.

Section function3.

Variables T1 T2 T3 : Type.

Lemma left_id_rec : forall (f : T1 -> T2) (Hbf : bijective f) x,
   ((inverse Hbf) \o f) x = x.
Proof.
move=> f [H1 H2] x; apply: (H1 ((inverse (conj H1 H2) \o f) x) x).
by rewrite -(pi2 (cid ((proj2 (conj H1 H2)) (f x)))).
Qed.

Lemma right_id_rec : forall (f : T1 -> T2) (Hbf : bijective f) y,
  (f \o (inverse Hbf)) y = y.
Proof.
move=> f Hbf y; rewrite (pi2 (cid ((proj2 Hbf) y))).
by apply: f_equal; rewrite -(pi2 (cid ((proj2 Hbf) y))).
Qed.

Lemma bij_rec : forall (f : T1 -> T2) (Hbf : bijective f),
  bijective (inverse Hbf).
Proof.
move=> f Hbf; split=> x.
  move=> y H.
  rewrite (pi2 (cid ((proj2 Hbf) y))).
  rewrite(pi2 (cid ((proj2 Hbf) x))).
  by apply: f_equal.
by exists (f x); rewrite -{1}(left_id_rec Hbf x).
Qed.

Lemma rec_invo : forall (f : T1 -> T2) (Hbf : bijective f) x,
   inverse (bij_rec Hbf) x = f x.
Proof.
move=> f Hbf x.
rewrite {2}(pi2 (cid ((proj2 (bij_rec Hbf)) x))).
by move: (right_id_rec Hbf); rewrite /funcomp => ->.
Qed.

End function3.

Section Family.

Variable T : Type. 
Variable Ti : eqType.

Structure family : Type := mkfamily  {
     ind :  Ti -> Prop;
     F :> Ti -> T -> Prop
   }.

Definition finite_set (T1 : eqType) (A : T1 -> Prop) :=
  exists s : seq T1, A =s (fun x => x \in s).    

Structure finite_family := Ff {
   fam :> family;
   _ : finite_set (ind fam)
}.

Lemma finfam : forall (f : finite_family), finite_set (ind f).
Proof. by case. Qed.

Definition ind_seq (f : finite_family) := pi1 (cid (finfam f)).

Lemma ind_seqP (f : finite_family) : (ind f) =s (fun x => x \in (ind_seq f)).
Proof. exact: (pi2 (cid (finfam f))). Qed.

Definition inter_fam (f : family) (x : T) := 
  forall i, i 'in (ind f)  -> x 'in (f i) .

Definition union_fam (f : family) (x : T) := 
  exists i, i 'in (ind f) /\ x 'in (f i).
 
Definition cover (A : T -> Prop) (f : family) := A <s union_fam f.

Definition sub_family f g := 
  ind g <s ind f /\ forall i, i 'in (ind g) -> f i =s g i.

Definition empty_fam (f : family) := ind f <s {ø : Ti}.

Lemma fam_empty (f : family) : (forall x : Ti, x 'in {ø : Ti}) ->  empty_fam f.
Proof. by move=> ?. Qed.

Lemma union_empty (f : family) : empty_fam f -> union_fam f =s {ø : T}.
Proof. by move=> Hf; split=> // [][y [/Hf H _]]. Qed.

Lemma inter_empty (f : family) : empty_fam f -> inter_fam f =s {full : T}.
Proof. by move=> Hf;split=> // Hx y /Hf. Qed.

Lemma in_union_fam (f : family) i : i 'in (ind f) -> f i <s union_fam f.
Proof. by move=> x Hx; exists i. Qed.

Lemma inter_fam_in (f : family) i : i 'in (ind f) -> inter_fam f <s f i.
Proof. by move=> H x; apply. Qed.

Definition cardinality (f : finite_family) := size (ind_seq f).

Lemma fempty_def f : empty_fam f -> finite_set (ind f).
Proof. by move=> Hx; exists [::];split=> // /Hx. Qed.

Definition fempty f (Hf : empty_fam f) := Ff (fempty_def Hf). 

Definition fam_fin_to_seq (f : finite_family) := map f (ind_seq f).

Lemma fam_card (f : finite_family) : size (fam_fin_to_seq f) = cardinality f.
Proof. by rewrite size_map. Qed.

Lemma fam_fin_v (f : finite_family) :
   (fam_fin_to_seq f) = [::] <-> empty_fam f.
Proof.
split=> [Hv x /ind_seqP Hx|Hv].
  have Hs : ind_seq f = [::].
    apply: size0nil.    
    have:= (fam_card f).
    by rewrite Hv /cardinality => <- .
  by rewrite Hs in Hx.
case Hs: (ind_seq f)=>[|s t]; first by apply: size0nil; rewrite size_map Hs. 
by have/Hv : s 'in (ind f) by apply/ind_seqP; rewrite Hs; apply: mem_head.
Qed.

Lemma card_empty f (H : empty_fam f) : cardinality (fempty H) = 0.
Proof.
rewrite -fam_card.
by have/fam_fin_v ->: empty_fam (fempty H).
Qed.

Canonical set_eqType := Eval hnf in EqType (T -> Prop) (T_eqMixin (T -> Prop)).

Lemma fam_fin_eq (f : finite_family) (g : T -> Prop) :
  g \in (fam_fin_to_seq f) <-> exists i : Ti, i 'in (ind f) /\ g = f i.
Proof.
split. 
  case: (classic (forall x : Ti, empty x)).
    move/(@fam_empty f)=> Hv.
    by have -> :  (fam_fin_to_seq f) = [::] by apply/fam_fin_v.
  case/not_all_ex_not=> x0 _.
  rewrite -index_mem size_map => H.   
  exists (nth x0 (ind_seq f) (index g (fam_fin_to_seq f))).
  split; first by apply/ind_seqP/mem_nth.
  rewrite -(@nth_map _ _ (T -> Prop) {ø : T} (F f)) // nth_index //.
  by rewrite -index_mem size_map.
case=> i [/ind_seqP H1 H2].
apply/(nthP {ø : T}).
exists (index i (ind_seq f)).
  by rewrite size_map index_mem.
by rewrite (nth_map i) // ?nth_index ?H2 // index_mem.
Qed.   

Definition inter_seq (s : seq (T -> Prop)) := 
  foldr (@inter T) {full : T} s.

Definition inter_fam_fin (f : finite_family) :=
  inter_seq (fam_fin_to_seq f).

Lemma in_inter_seq : forall (s : seq (T -> Prop)) g, g \in s -> 
  inter_seq s <s g.
Proof.
elim=> //.
move=> A s HR g /=; rewrite in_cons; case/orP=>[/eqP ->|Hg].
  exact: inter_left.
by apply/(include_trans (@inter_right _ A (inter_seq s)))/HR.
Qed.

Lemma in_inter_seq2 : forall (s : seq (T -> Prop)) x, 
  (forall g, g \in s -> x 'in g) -> x 'in (inter_seq s).
Proof.
elim=> //= A s HR x Hx; split.
  by apply/Hx/mem_head.
by apply:HR => g Hg; apply/Hx/mem_behead.
Qed.

Lemma inter_fin_eq (f : finite_family) : inter_fam f =s inter_fam_fin f.
Proof.
split=> [Hx|Hx y Hy].
  by apply: in_inter_seq2=> g; case/fam_fin_eq=> y [Hy ->]; apply: inter_fam_in.
have/in_inter_seq : (f y) \in (fam_fin_to_seq f) by apply/fam_fin_eq; exists y.
by apply.
Qed.

End Family.
